import numpy as np
import pysa.emd as emd
import pysa.nhht as nhht
import pysa.utils as utils
import util.file_handler as file_handler
import sys


########################################################################################################################
#                                  Activity: Adaptive signal analysis using NHHT                                       #
#                                                                                                                      #
#   This activity loads the EMD-filtered EEG data and the bad trials obtained from 'activity_clean_eeg.py'             #
#   and applies the Normalized Hilbert-Huang Transform (NHHT). The frequencies and amplitudes for the first four IMFs  #
#   are stored for each electrode, for each stimulus type (condition). That is, the resulting data matrix will have    #
#   dimensions = [n_electrodes x n_imfs x data_length]. (Typically: [16, 4, 500000])                                   #
#                                                                                                                      #
########################################################################################################################

def start(username=''):

    # --------------------------------------USER INFORMATION AND PARAMETERS------------------------------------------- #
    if username == '':
        print "Please specify subject username in 'main.py'"
        sys.exit()
    sample_frequency = 500.0  # [Hz]
    samples_per_trial = 3000  # 6sec trial length

    # ---------------------------------------------------------------------------------------------------------------- #

    # -------------------------------------------FILENAMES------------------------------------------------------------ #
    filename_seg = username + '_data_motor_cortex_segmented.mat'
    filename_filtered_data = 'data_matrix_emd_filtered.mat'
    # ---------------------------------------------------------------------------------------------------------------- #

    # ------------------------------------------RAW EEG DATA---------------------------------------------------------- #
    data_json = file_handler.read_mat_file_into_json_object(username, filename_seg)
    events = data_json['events']
    # ---------------------------------------------------------------------------------------------------------------- #

    # ------------------------------------------EMD-FILTERED DATA----------------------------------------------------- #
    data_json = file_handler.read_mat_file_into_json_object(username, filename_filtered_data)
    data_matrix_filtered = data_json[filename_filtered_data[:-4]]
    # ---------------------------------------------------------------------------------------------------------------- #

    # ---------------------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------------------------------------------------------------------------- #
    # ---------------------------------------------------------------------------------------------------------------- #

    # ---------------------------------------------------------------------------------------------------------------- #
    #                                                DATA ANALYSIS                                                     #
    # ---------------------------------------------------------------------------------------------------------------- #
    print("\x1b[2J\x1b[H")
    introduction = "EEG adaptive signal analysis (NHT) - Subject = " + username.title()
    print "%s\n%s\n%s" % ('-' * len(introduction), introduction, '-' * len(introduction))
    # PARAMETERS
    max_modes = 5
    max_siftings = 400
    sample_frequency = int(sample_frequency)

    # Define which IMFs to store (First four IMFs)
    imf_start = 0
    imf_end = 4
    n_imfs = imf_end - imf_start
    # Number of electrodes:
    n_electrodes = len(data_matrix_filtered)

    final_trial_length = 2250  # = 4,5 s (0.5s before stimulus onset, 3s movement, and 1s after movement)
    n_trials = len(events)
    n_left = 0
    n_right = 0
    n_neutral = 0
    for trial in range(n_trials):
        if events[trial] == 'L':
            n_left += 1
        elif events[trial] == 'R':
            n_right += 1
        elif events[trial] == 'N':
            n_neutral += 1

    frequency_matrix_left = np.zeros((n_electrodes, n_imfs, final_trial_length*n_left))
    amplitude_matrix_left = np.zeros((n_electrodes, n_imfs, final_trial_length*n_left))

    frequency_matrix_right = np.zeros((n_electrodes, n_imfs, final_trial_length*n_right))
    amplitude_matrix_right = np.zeros((n_electrodes, n_imfs, final_trial_length*n_right))

    frequency_matrix_neutral = np.zeros((n_electrodes, n_imfs, final_trial_length*n_neutral))
    amplitude_matrix_neutral = np.zeros((n_electrodes, n_imfs, final_trial_length*n_neutral))

    print "Starting EMD and NHHT analysis"
    for el in range(n_electrodes):
        first_start_cut = 100
        first_end_cut = first_start_cut + 2700
        second_start_cut = 150
        second_end_cut = second_start_cut + 2250

        left_counter = 0
        right_counter = 0
        neutral_counter = 0

        # EMD on each trial of the current electrode
        for trial in range(n_trials):

            start_sample = (trial*samples_per_trial)
            end_sample = start_sample + samples_per_trial

            data = data_matrix_filtered[el, start_sample:end_sample]
            min_data = min(data)
            max_data = max(data)

            # EMD:
            imfs = emd.emd(data, min_data, max_data, max_modes=max_modes, max_siftings=max_siftings)
            imfs = imfs[imf_start:imf_end]
            # First cut to get rid of EMD end-effects
            imfs = imfs[:, first_start_cut:first_end_cut]
            # Scaling --> get instantaneous amplitude
            scaled_imfs, inst_amp = utils.scale_imfs_amplitude(imfs)
            # Reverse normalize amplitude
            for i in range(len(inst_amp)):
                inst_amp[i] = utils.reverse_normalization(inst_amp[i], min_data, max_data, len(inst_amp[i]))
            # NHHT --> get instantaneous frequency
            inst_freq = nhht.get_instantaneous_frequency(scaled_imfs, sample_frequency)
            # Second cut to get rid of Hilbert-transform end-effects
            inst_freq = inst_freq[:, second_start_cut:second_end_cut]
            inst_amp = inst_amp[:, second_start_cut:second_end_cut]

            # Check whether the current stimulus is LEFT, RIGHT, or NEUTRAL
            if events[trial] == 'L':
                # Store the instantaneous frequencies and amplitudes
                frequency_matrix_left[el, :, left_counter*final_trial_length:(left_counter+1)*final_trial_length] = inst_freq
                amplitude_matrix_left[el, :, left_counter*final_trial_length:(left_counter+1)*final_trial_length] = inst_amp
                left_counter += 1

            elif events[trial] == 'R':
                # Store the instantaneous frequencies and amplitudes
                frequency_matrix_right[el, :, right_counter*final_trial_length:(right_counter+1)*final_trial_length] = inst_freq
                amplitude_matrix_right[el, :, right_counter*final_trial_length:(right_counter+1)*final_trial_length] = inst_amp
                right_counter += 1

            elif events[trial] == 'N':
                # Store the instantaneous frequencies and amplitudes
                frequency_matrix_neutral[el, :, neutral_counter*final_trial_length:(neutral_counter+1)*final_trial_length] = inst_freq
                amplitude_matrix_neutral[el, :, neutral_counter*final_trial_length:(neutral_counter+1)*final_trial_length] = inst_amp
                neutral_counter += 1

        elec = el + 1
        if elec == int(0.1*n_electrodes):
            print 'Normalized Hilbert transform procedure: 10% finished...'
        elif elec == int(0.2*n_electrodes):
            print 'Normalized Hilbert transform procedure: 20% finished...'
        elif elec == int(0.3*n_electrodes):
            print 'Normalized Hilbert transform procedure: 30% finished...'
        elif elec == int(0.4*n_electrodes):
            print 'Normalized Hilbert transform procedure: 40% finished...'
        elif elec == int(0.5*n_electrodes):
            print 'Normalized Hilbert transform procedure: 50% finished...'
        elif elec == int(0.6*n_electrodes):
            print 'Normalized Hilbert transform procedure: 60% finished...'
        elif elec == int(0.7*n_electrodes):
            print 'Normalized Hilbert transform procedure: 70% finished...'
        elif elec == int(0.8*n_electrodes):
            print 'Normalized Hilbert transform procedure: 80% finished...'
        elif elec == int(0.9*n_electrodes):
            print 'Normalized Hilbert transform procedure: 90% finished...'

    print "Finished! Storing all the data."

    print np.shape(frequency_matrix_left)
    print np.shape(frequency_matrix_right)
    print np.shape(frequency_matrix_neutral)

    subfolder = 'NHHT/'
    fn_left = 'nht_left_hand.mat'
    fn_right = 'nht_right_hand.mat'
    fn_neutral = 'nht_neutral.mat'

    file_handler.save_hht_data_as_mat(frequency_matrix_left, amplitude_matrix_left, username, subfolder, fn_left)
    file_handler.save_hht_data_as_mat(frequency_matrix_right, amplitude_matrix_right, username, subfolder, fn_right)
    file_handler.save_hht_data_as_mat(frequency_matrix_neutral, amplitude_matrix_neutral, username, subfolder,
                                      fn_neutral)









